"""Utilities for handling a working directory
   A work direction must have one and only one prjfile and a list of files,
   a WorkDir_Error will be raised if none or duplicate .prj files are found
"""

import md5
import os
import fnmatch
import pickle
import string
import time
import stat

class DirSupport:

    """Provide some utilities for handling dirs and files

    Based on RCSProxy.py and FSProxy.py in python1.5.2
    """

    def __init__(self):
        self._dirstack = []
	self.dir = os.curdir
	
    def __del__(self):
        self._close()

    def _close(self):
        while self._dirstack:
            self.back()

    def _addfile(self, fn):
	self._FILES.append(fn)

    def _filter(self, files, pats = [], igns = []):
	if pats:
	    def accept(name, pats = pats):
		for pat in pats:
		    if fnmatch.fnmatch(name, pat): return 1
		return 0
	    files = filter(accept, files)
	if igns:
	    def reject(name, igns = igns):
		for ign in igns:
		    if fnmatch.fnmatch(name, ign): return 0
		return 1
	    files = filter(reject, files)
	files.sort()
	return files

    def _visit(self, arg, dir, files):
	# perform action when cond is satisfied or None after filtering 
	action, cond, pats, igns = arg
	files = self._filter(files, pats, igns)
	for f in files:
	    file = (dir + '/' + f)[2:]
	    if cond == None or cond(file):
		action(file)

    def pwd(self):
        return os.getcwd()

    def cd(self, name):
        save = os.getcwd()
        os.chdir(name)
        self._dirstack.append(save)

    def back(self):
        if not self._dirstack:
            raise os.error, "empty directory stack"
        dir = self._dirstack[-1]
        os.chdir(dir)
        del self._dirstack[-1]

    def listall(self, pats = [], igns = []):
	files = os.listdir(os.curdir)
	return self._filter(files, pats, igns)

    def listvalid(self, pats = [], igns = []):
	files = os.listdir(os.curdir)
	files = filter(self.isvalid, files)
	return self._filter(files, pats, igns)

    def listfiles(self, pats = [], igns = []):
        files = os.listdir(os.curdir)
        files = filter(os.path.isfile, files)
        return self._filter(files, pats, igns)

    def listsubdirs(self, pats = [], igns = []):
        files = os.listdir(os.curdir)
        files = filter(self.isdir, files)
        return self._filter(files, pats, igns)

    def recursivelistall(self, pats = [], igns = []):
	self._FILES = []
	os.path.walk('.', self._visit, (self._addfile, None, pats, igns))
	return self._FILES

    def recursivelistvalid(self, pats = [], igns = []):
	self._FILES = []
	os.path.walk('.', self._visit, (self._addfile, self.isvalid, pats, igns))
	return self._FILES

    def recursivelistfiles(self, pats = [], igns = []):
	self._FILES = []
	os.path.walk('.', self._visit, (self._addfile, self.isfile, pats, igns))
	return self._FILES

    def recursivelistsubdirs(self, pats = [], igns = []):
	self._FILES = []
	os.path.walk('.', self._visit, (self._addfile, self.isdir, pats, igns))
	return self._FILES

    def isdir(self, name):
        return os.path.isdir(name)

    def isfile(self, name):
	return os.path.isfile(name)
    
    def isvalid(self, name):
	return os.path.isfile(name) or os.path.isdir(name)

    def islink(self, name):
	return os.path.islink(name)

    def mkdir(self, name):
        os.mkdir(name, 0777)

    def rmdir(self, name):
        os.rmdir(name)

    def rmfile(self, name):
	return os.unlink(name)

    def readlink(self, name):
	return os.readlink(name)

    def mksymlink(self, name, sym):
	return os.symlink(name, sym)

    def sum(self, file):
	# return None for zero-length file, regard zero-length file as
	# if nonexistent.
	if os.stat(file)[-4] == 0:
	    return None
        f = open(file)
        BUFFERSIZE = 1024*8
        sum = md5.new()
        while 1:
            buffer = f.read(BUFFERSIZE)
            if not buffer:
                break
            sum.update(buffer)
        return hexify(sum.digest())
	
    def unkeysum(self, file):
	pass

    def stamp(self, file):
	st = os.stat(file)
	mode = stat.S_IMODE(st[stat.ST_MODE])
	atime, mtime = st[stat.ST_ATIME], st[stat.ST_MTIME]
	return (mode, atime, mtime, time.asctime(time.gmtime(mtime)))

    def setstamp(self, file, mode, atime, mtime):
	os.chmod(file, mode)
	os.utime(file, (atime, mtime))

    def backup(self, file):
	if os.path.isfile(file):
	    bfile = file + '~'
	    try: os.unlink(bfile)
	    except os.error: pass
	    os.rename(file, bfile)

	
class WorkDir(DirSupport):

    """ Provide a object representing the working dir

    Variables:
    dict: a dictionary stores each file in the directory
          file: ('F', mtime, md5sum) for a file

    """

    IgnoreList = ['.*', 'core', '@*', ',*', '#*#', '*~', '*.o', '*.a', '*.so', '*.pyc']
    WDSTATE = '.wd_state'
    
    def __init__(self):
	DirSupport.__init__(self)
        self.dict = {}

    def _action(self, file):
	if self.isfile(file):
	    self.dict[file] = (os.stat(file)[-2], self.sum(file))

    def buildsum(self):
	if os.path.exists(self.WDSTATE):
	    self.dict = pickle.load(open(self.WDSTATE))
	else:
	    self.dict = {}
	    os.path.walk('.', self._visit, (self._action, None, [], self.IgnoreList))
	    pickle.dump(self.dict, open(self.WDSTATE, 'w'))
	
    def md5(self, file):
	"""a fast md5sum of a file, only calculated again if mtime is newer """
	ntime = os.stat(file)[-2]
	try:
	    otime, osum = self.dict[file]
	    if ntime > otime:
		nsum = self.sum(file)
		self.dict[file] = (ntime, nsum)
		pickle.dump(self.dict, open(self.WDSTATE, 'w'))
		return nsum
	    else:
		return osum
	except KeyError:
	    nsum = self.sum(file)
	    self.dict[file] = (ntime, nsum)
	    pickle.dump(self.dict, open(self.WDSTATE, 'w'))
	    return nsum

    def clean(self, pats = []):
	if not pats:
	    pats=self.IgnoreList
	for f in DirSupport.recursivelistfiles(self, pats, ['.*', '*.a', '*.so', '*~']):
	    self.rmfile(f)
	
    def workfiles(self):
	return self.recursivelistfiles([], self.IgnoreList)
	
    def status(self, files = []):
	if not files:
	    files = self.workfiles()
	rept = {}
	for f in files:
	    try:
		rept[f] = self.md5(f)
	    except OSError:
		rept[f] = None
	return rept
    

class RWorkDir(WorkDir):

    """Local Working Directory with proxy"""

    def __init__(self):
	WorkDir.__init__(self)
	self.proxy = None
	
    def setproxy(self, proxy):
	self.proxy = proxy
	
    def check(self, files = []):

	"""Report on a list of files' status:
	
	status:  Modified, Uptodate, Added and Removed
	
	Status:
	'???' -- Unknown, not in local or remote working direction
	'Omt' -- Missing
	'New' -- new local file
	'Mod' -- Changed locally
	'===' -- Uptodate, everything's fine

	"""

	local = self.status()
	remote = self.proxy.status()
	lfiles = local.keys()
	rfiles = remote.keys()
	dict = {}
	
	if not files:
	    for f in lfiles:
		if f in remote.keys():
		    dict[f] = local[f] == remote[f] and '===' or 'Mod'
		    if rfiles: rfiles.remove(f)
		else:
		    dict[f] = 'New'
	    for f in rfiles:
		dict[f] = 'Rmv'

	else:
	    for f in files:
		if f in lfiles and f in rfiles:
		    dict[f] = local[f] == remote[f] and '===' or 'Mod'
		elif f in lfiles:
		    dict[f] = 'New'
		else:
		    dict[f] = 'Rmv'

	return dict
    


# hexify and unhexify are useful to print MD5 checksums in hex format

hexify_format = '%02x' * 16
def hexify(sum):
        "Return a hex representation of a 16-byte string (e.g. an MD5 digest)"
        if sum is None:
                return "None"
        return hexify_format % tuple(map(ord, sum))

def unhexify(hexsum):
        "Return the original from a hexified string"
        if hexsum == "None":
                return None
        sum = ''
        for i in range(0, len(hexsum), 2):
                sum = sum + chr(string.atoi(hexsum[i:i+2], 16))
        return sum







